"""Manage configuration parameters."""

import enum
import pathlib
import typing

import aiofiles
import pydantic
import ruamel.yaml

DEFAULT_CONFIGURATION_FILE = pathlib.Path(".pygitsync.yaml")
DEFAULT_SLEEP_INTERVAL_SECONDS = 120.0
EXCEPTION_SLEEP_DURATION_SECONDS = 15.0


class GitRefType(str, enum.Enum):
    """Type of git reference to sync against."""

    branch = "branch"
    tag = "tag"


class ApplicationConfiguration(pydantic.BaseModel):
    """General application configuration."""

    exception_sleep_seconds: float = EXCEPTION_SLEEP_DURATION_SECONDS
    is_daemon: bool = False
    sleep_interval_seconds: float = DEFAULT_SLEEP_INTERVAL_SECONDS


class RepoConfiguration(pydantic.BaseModel):
    """Configuration of git repository for synchronization."""

    pattern_type: GitRefType
    pattern: str
    url: str

    class Config:
        """Pydantic model configuration."""

        use_enum_values = True


class GitSyncConfiguration(pydantic.BaseModel):
    """Collector of all configuration parameters."""

    repo: RepoConfiguration

    application: ApplicationConfiguration = ApplicationConfiguration()


async def _load_configuration(
    configuration_file: pathlib.Path,
    is_daemon: bool,
    sleep_interval_seconds: typing.Optional[float],
) -> GitSyncConfiguration:
    """
    Acquire configuration values from storage.

    Args:
        configuration_file: Path to the expected configuration file.
        sleep_interval_seconds: User defined sleep interval between git sync
        polling.

    Returns:
        Configuration values acquired.
    """
    async with aiofiles.open(configuration_file, mode="r") as f:
        content = await f.read()

    yaml_parser = ruamel.yaml.YAML(typ="safe")
    yaml_data = yaml_parser.load(content)
    value = GitSyncConfiguration.parse_obj(yaml_data)

    value.application.is_daemon = is_daemon
    if sleep_interval_seconds:
        value.application.sleep_interval_seconds = sleep_interval_seconds

    return value



import logging

import pytest

import pygitsync._repo
from pygitsync._repo import (
    GitRefType,
    InvalidBranchError,
    RepoConfiguration,
    _evaluate_branch,
    _evaluate_tag,
    _fetch_repo,
)

from .repo_support import (
    DefaultBranchHead,
    FreshClone,
    MockRepo,
    NonDefaultBranchHead,
    RepoSupport,
    TaggedRepo,
)


@pytest.fixture()
def mock_configuration(mocker):
    rc = RepoConfiguration(
        pattern_type=GitRefType.branch,
        pattern="some_branch",
        url="user@some.where:this/project",
    )
    return rc


@pytest.fixture()
def mock_get_remote(mocker):
    return mocker.patch(
        "pygitsync._repo._get_remote", return_value=mocker.MagicMock()
    )


@pytest.fixture()
def mock_repo(mocker):
    def _apply(
        repo_url: str, repo_support: RepoSupport, initial_branch: str
    ) -> MockRepo:
        return MockRepo(repo_url, repo_support, initial_branch)

    return _apply


class TestEvaluateBranch:
    def test_default_branch(
        self, mock_configuration, mock_get_remote, mock_repo
    ):
        this_pattern = r"main"
        with mock_repo(
            mock_configuration.url,
            DefaultBranchHead(),
            initial_branch=this_pattern,
        ).instance() as repo_instance:
            this_repo = repo_instance.repo

            _evaluate_branch(this_pattern, this_repo)

            assert this_repo
            assert not this_repo.head.is_detached
            assert this_repo.head.reference == this_repo.heads.main
            # ensure files are checked out in filesystem
            assert (repo_instance.repo_dir / "c2").is_file()

    def test_nondefault_branch(
        self, mock_configuration, mock_get_remote, mock_repo
    ):
        this_pattern = r"feature"
        with mock_repo(
            mock_configuration.url,
            NonDefaultBranchHead(),
            initial_branch="main",
        ).instance() as repo_instance:
            this_repo = repo_instance.repo

            _evaluate_branch(this_pattern, this_repo)

            assert this_repo
            assert not this_repo.head.is_detached
            assert this_repo.head.reference == this_repo.heads.feature
            # ensure files are checked out in filesystem
            assert (repo_instance.repo_dir / "fc1").is_file()

    def test_branch_checkout(
        self, mock_configuration, mock_get_remote, mock_repo
    ):
        """Mismatched branch checks out necessary branch."""
        this_pattern = r"feature"
        with mock_repo(
            mock_configuration.url,
            NonDefaultBranchHead(),
            initial_branch="main",
        ).instance() as repo_instance:
            this_repo = repo_instance.repo
            this_repo.head.reference = this_repo.heads.main
            this_repo.head.reset(index=True, working_tree=True)

            assert (repo_instance.repo_dir / "dc2").is_file()

            _evaluate_branch(this_pattern, this_repo)

            assert this_repo
            assert not this_repo.head.is_detached
            assert this_repo.head.reference == this_repo.heads.feature
            # ensure files are checked out in filesystem
            assert (repo_instance.repo_dir / "fc1").is_file()

    def test_nonexistent_branch_raises(
        self, mock_configuration, mock_get_remote, mock_repo
    ):
        with mock_repo(
            mock_configuration.url,
            NonDefaultBranchHead(),
            initial_branch="master",
        ).instance() as repo_instance:
            with pytest.raises(
                InvalidBranchError, match=r"^Invalid branch name"
            ):
                this_repo = repo_instance.repo

                _evaluate_branch("main", this_repo)


class TestEvaluateTag:
    def test_no_tags_found(
        self, caplog, mock_configuration, mock_get_remote, mock_repo
    ):
        """No tags matching pattern silently exits with a warning."""
        this_pattern = r"^4\.6"
        with caplog.at_level(logging.WARNING):
            with mock_repo(
                mock_configuration.url,
                TaggedRepo(["3.1.4", "v3.2", "v3.1.5"]),
                initial_branch="main",
            ).instance() as repo_instance:
                _evaluate_tag(this_pattern, repo_instance.repo)

            captured_text = caplog.text
            assert "No tags found matching pattern" in captured_text

    def test_no_tags_in_repo(
        self, caplog, mock_configuration, mock_get_remote, mock_repo
    ):
        """No tags in repo silently exits with a warning."""
        this_pattern = r"^(v)?\d+\.\d+\.\d+"
        with caplog.at_level(logging.WARNING):
            with mock_repo(
                mock_configuration.url,
                DefaultBranchHead(),
                initial_branch="main",
            ).instance() as repo_instance:
                _evaluate_tag(this_pattern, repo_instance.repo)

            captured_text = caplog.text
            assert "No tags found matching pattern" in captured_text

    def test_single_tag(self, mock_configuration, mock_get_remote, mock_repo):
        this_pattern = r"^(v)?\d+\.\d+\.\d+"
        with mock_repo(
            mock_configuration.url, TaggedRepo(["3.1.4"]), initial_branch="main"
        ).instance() as repo_instance:
            this_repo = repo_instance.repo

            _evaluate_tag(this_pattern, this_repo)

            assert this_repo
            assert not this_repo.head.is_detached
            assert this_repo.head.reference == this_repo.tags[0]
            assert this_repo.tags[0].name == "3.1.4"
            # check for the file in the commit that the tag is on
            assert (repo_instance.repo_dir / "c0").is_file()

    def test_tags_release_order(
        self, mock_configuration, mock_get_remote, mock_repo
    ):
        """Semantic release reasoning defines the "latest" tag found, not tag date."""
        this_pattern = r"^(v)?\d+(\.\d+)*"
        with mock_repo(
            mock_configuration.url,
            TaggedRepo(["3.1.4", "v3.2", "v3.1.5"]),
            initial_branch="main",
        ).instance() as repo_instance:
            this_repo = repo_instance.repo

            _evaluate_tag(this_pattern, this_repo)

            assert this_repo
            assert not this_repo.head.is_detached
            assert this_repo.head.commit == this_repo.commit("v3.2")
            assert this_repo.head.reference.name == "v3.2"
            # check for the file in the commit that the tag is on
            assert (repo_instance.repo_dir / "c1").is_file()


class TestFetchRepo:
    @pytest.mark.asyncio
    async def test_new_repo_created(
        self, caplog, mock_configuration, mock_get_remote, mock_repo, mocker
    ):
        """A repo that has not been cloned before is cloned/created."""
        mock_configuration.pattern = r"main"
        mock_configuration.pattern_type = GitRefType.branch
        with mock_repo(
            mock_configuration.url, DefaultBranchHead(), initial_branch="main"
        ).instance() as remote_instance:
            with mock_repo(
                str(remote_instance.repo_dir),
                FreshClone(mocker, remote_instance),
                initial_branch="main",
            ).instance() as repo_instance:
                assert not repo_instance.repo_dir.exists()

                await _fetch_repo(mock_configuration, repo_instance.working_dir)

                assert repo_instance.repo_dir.is_dir()

    @pytest.mark.asyncio
    async def test_branch_head_rule(
        self, caplog, mock_configuration, mock_get_remote, mock_repo, mocker
    ):
        b = mocker.MagicMock()
        t = mocker.MagicMock()
        pygitsync._repo.GIT_REF_METHOD = {
            GitRefType.branch: b,
            GitRefType.tag: t,
        }
        mock_configuration.pattern = "main"
        mock_configuration.pattern_type = GitRefType.branch
        with mock_repo(
            mock_configuration.url, DefaultBranchHead(), initial_branch="main"
        ).instance() as repo_instance:
            await _fetch_repo(mock_configuration, repo_instance.working_dir)

        t.assert_not_called()
        b.assert_called_once_with(mock_configuration.pattern, mocker.ANY)

    @pytest.mark.asyncio
    async def test_tag_pattern_rule(
        self, caplog, mock_configuration, mock_get_remote, mock_repo, mocker
    ):
        b = mocker.MagicMock()
        t = mocker.MagicMock()
        pygitsync._repo.GIT_REF_METHOD = {
            GitRefType.branch: b,
            GitRefType.tag: t,
        }
        mock_configuration.pattern = r"^(v|V)?\d+\.\d+\.\d+"
        mock_configuration.pattern_type = GitRefType.tag
        with mock_repo(
            mock_configuration.url,
            TaggedRepo(["3.1.2", "v3.4"]),
            initial_branch="main",
        ).instance() as repo_instance:
            await _fetch_repo(mock_configuration, repo_instance.working_dir)

        b.assert_not_called()
        t.assert_called_once_with(mock_configuration.pattern, mocker.ANY)



import pathlib
import typing

import pytest
from click.testing import CliRunner

from pygitsync._configuration import (
    ApplicationConfiguration,
    GitSyncConfiguration,
    RepoConfiguration,
)
from pygitsync.entrypoint import (
    WhileContext,
    flit_entry,
    main,
    process_cli_arguments,
)


@pytest.fixture()
def mock_configuration(mocker):
    value = GitSyncConfiguration.parse_obj(
        {
            "application": ApplicationConfiguration.parse_obj(
                {
                    "exception_sleep_seconds": 0.1,
                    "sleep_duration_seconds": 0.1,
                }
            ),
            "repo": RepoConfiguration.parse_obj(
                {
                    "pattern": "master",
                    "pattern_type": "branch",
                    "url": "https://some.where",
                }
            ),
        }
    )
    return mocker.patch(
        "pygitsync.entrypoint._load_configuration", return_value=value
    )


@pytest.fixture()
def mock_daemon_context(mocker):
    return mocker.patch("pygitsync.entrypoint.daemon.DaemonContext")


@pytest.fixture()
def mock_main(mocker):
    return mocker.patch("pygitsync.entrypoint.main", mocker.AsyncMock())


@pytest.fixture()
def mock_runner() -> CliRunner:
    return CliRunner()


@pytest.fixture()
def mock_while_context(mocker):
    def _apply(returns: typing.List[bool]):
        import pygitsync.entrypoint

        mocker.patch(
            "pygitsync._configuration.EXCEPTION_SLEEP_DURATION_SECONDS", 0.1
        )

        return mocker.patch.object(
            pygitsync.entrypoint.WhileContext,
            "keep_running",
            side_effect=returns,
        )

    return _apply


class TestWhileContext:
    def test_is_daemon(self):
        under_test = WhileContext(True)

        assert under_test.is_daemon

        # keep_running is true both times (looks like redundant tests, but it's not)
        assert under_test.keep_running()
        assert under_test.keep_running()

    def test_not_daemon(self):
        under_test = WhileContext(False)

        assert not under_test.is_daemon

        assert under_test.keep_running()
        assert not under_test.keep_running()


class TestMain:
    MOCK_PATH = pathlib.Path("some/path")

    @pytest.mark.asyncio
    async def test_is_daemon(
        self, mock_configuration, mock_while_context, mocker
    ):
        # cannot test true infinite loop, but at least show that it doesn't
        # exit after the first iteration.
        mock_keep_running = mock_while_context([True, True, False])
        await main(self.MOCK_PATH, True, 1)

        mock_keep_running.assert_has_calls(
            [
                mocker.call(),
                mocker.call(),
                mocker.call(),
            ]
        )

    @pytest.mark.asyncio
    async def test_not_daemon(
        self, mock_configuration, mock_while_context, mocker
    ):
        mock_keep_running = mock_while_context([True, False])
        await main(self.MOCK_PATH, True, 1)

        mock_keep_running.assert_has_calls(
            [
                mocker.call(),
                mocker.call(),
            ]
        )


class TestProcessCliArguments:
    def test_default(
        self, mock_configuration, mock_daemon_context, mock_main, mock_runner
    ):
        result = mock_runner.invoke(process_cli_arguments, [])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path(".pygitsync.yaml"), False, None
        )
        mock_daemon_context.assert_not_called()

    def test_configuration_long(
        self, mock_configuration, mock_daemon_context, mock_main, mock_runner
    ):
        result = mock_runner.invoke(
            process_cli_arguments, ["--configuration", "some/file"]
        )

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path("some/file"), False, None
        )
        mock_daemon_context.assert_not_called()

    def test_configuration_short(
        self, mock_configuration, mock_daemon_context, mock_main, mock_runner
    ):
        result = mock_runner.invoke(process_cli_arguments, ["-c", "some/file"])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path("some/file"), False, None
        )
        mock_daemon_context.assert_not_called()

    def test_interval_long(self, mock_configuration, mock_main, mock_runner):
        result = mock_runner.invoke(process_cli_arguments, ["--interval", "25"])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path(".pygitsync.yaml"), False, 25
        )

    def test_interval_short(self, mock_configuration, mock_main, mock_runner):
        result = mock_runner.invoke(process_cli_arguments, ["-i", "25"])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path(".pygitsync.yaml"), False, 25
        )

    def test_enable_daemon_long(
        self, mock_configuration, mock_daemon_context, mock_main, mock_runner
    ):
        result = mock_runner.invoke(process_cli_arguments, ["--daemon"])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path(".pygitsync.yaml"), True, None
        )
        mock_daemon_context.assert_called_once()

    def test_enable_daemon_short(
        self, mock_configuration, mock_daemon_context, mock_main, mock_runner
    ):
        result = mock_runner.invoke(process_cli_arguments, ["-d"])

        assert result.exit_code == 0
        mock_main.assert_awaited_once_with(
            pathlib.Path(".pygitsync.yaml"), True, None
        )
        mock_daemon_context.assert_called_once()


class TestFlitEntry:
    def test_clean(self, mocker):
        mock_main = mocker.patch("pygitsync.entrypoint.process_cli_arguments")

        flit_entry()

        mock_main.assert_called_once()



import pathlib
import tempfile

import pytest

from pygitsync._configuration import (
    GitRefType,
    RepoConfiguration,
    _load_configuration,
)


@pytest.fixture()
def generate_config_file():
    def _apply(content: str, file_path: pathlib.Path) -> None:
        with file_path.open(mode="w") as f:
            f.write(content)

    return _apply


class TestLoadConfiguration:
    MOCK_INTERVAL = 4.56

    @pytest.mark.asyncio
    async def test_clean(self, generate_config_file):
        expected = {
            "daemon": True,
            "exception": 6.28,
            "sleep": 3.14,
            "pattern": "master",
            "pattern_type": GitRefType.branch,
            "url": "https://some.where/git/repo.git",
        }
        ingress_content = f"""---
application:
  exception_sleep_seconds: {expected["exception"]}
  is_daemon: {"true" if expected["daemon"] else "false"}
  sleep_interval_seconds: {expected["sleep"]}

repo:
  pattern_type: {expected["pattern_type"].name}
  pattern: {expected["pattern"]}
  url: {expected["url"]}
"""

        with tempfile.TemporaryDirectory() as d:
            working_dir = pathlib.Path(d)
            expected_path = working_dir / "some_config.yaml"

            generate_config_file(ingress_content, expected_path)

            result = await _load_configuration(
                expected_path,
                expected["daemon"],
                None,
            )

            assert (
                result.application.exception_sleep_seconds
                == expected["exception"]
            )
            assert result.application.is_daemon == expected["daemon"]
            assert (
                result.application.sleep_interval_seconds == expected["sleep"]
            )
            assert result.repo.pattern == expected["pattern"]
            assert result.repo.pattern_type == expected["pattern_type"]
            assert result.repo.url == expected["url"]

    @pytest.mark.asyncio
    async def test_sleep_cli(self, generate_config_file):
        """Sleep duration function argument takes precedence over yaml."""
        expected = {
            "daemon": True,
            "exception": 6.28,
            "sleep": 3.14,
            "pattern": "master",
            "pattern_type": GitRefType.branch,
            "url": "https://some.where/git/repo.git",
        }
        ingress_content = f"""---
application:
  exception_sleep_seconds: {expected["exception"]}
  is_daemon: {"true" if expected["daemon"] else "false"}
  sleep_interval_seconds: {self.MOCK_INTERVAL}

repo:
  pattern_type: {expected["pattern_type"].name}
  pattern: {expected["pattern"]}
  url: {expected["url"]}
"""

        with tempfile.TemporaryDirectory() as d:
            working_dir = pathlib.Path(d)
            expected_path = working_dir / "some_config.yaml"

            generate_config_file(ingress_content, expected_path)

            result = await _load_configuration(
                expected_path,
                expected["daemon"],
                expected["sleep"],
            )

            assert (
                result.application.exception_sleep_seconds
                == expected["exception"]
            )
            assert result.application.is_daemon == expected["daemon"]
            assert (
                result.application.sleep_interval_seconds == expected["sleep"]
            )
            assert result.repo.pattern == expected["pattern"]
            assert result.repo.pattern_type == expected["pattern_type"]
            assert result.repo.url == expected["url"]



import logging
import pathlib
import re
from urllib.parse import unquote, urlparse

log = logging.getLogger(__name__)

SCP_STYLE_REGEX = r"^(\w+@)?[\w.]+:[\w/]+"
URL_PREFIX_REGEX = r"^(file)|(ftp(s)?)|(git)|(http(s)?)|(ssh)://"


def _extract_from_url(url: str) -> str:
    return (
        pathlib.PurePosixPath(unquote(urlparse(url).path))
        .parts[-1]
        .replace(".git", "")
    )


def _extract_from_raw_path(path: str) -> str:
    return pathlib.PurePosixPath(path).parts[-1].replace(".git", "")


def _extract_from_scp_syntax(url: str) -> str:
    tokens = url.split(":")

    return (
        pathlib.PurePosixPath(unquote(tokens[1])).parts[-1].replace(".git", "")
    )


def extract_git_project(url: str) -> str:
    """Extract git project name from a git repo URL.

    Args:
        url: Git repo URL to parse

    Returns:
        Git project name
    """
    if re.match(URL_PREFIX_REGEX, url.lower()):
        log.debug(f"extracting git project name from URL, {url}")
        value = _extract_from_url(url)
    elif re.match(SCP_STYLE_REGEX, url.lower()):
        log.debug(
            f"extracting git project name from SCP-style path-spec, {url}"
        )
        value = _extract_from_scp_syntax(url)
    else:
        log.debug(f"extracting git project name from raw posix path, {url}")
        value = _extract_from_raw_path(url)

    return value



import abc
import contextlib
import pathlib
import tempfile
import typing

import git

import pygitsync._repo
from pygitsync._url import extract_git_project


class RepoSupport(metaclass=abc.ABCMeta):
    repo: typing.Optional[git.Repo]

    def __init__(self):
        self.repo = None

    def _commit_file(self, file_path: pathlib.Path) -> None:
        assert self.repo

        filename = str(file_path)
        open(filename, "wb").close()
        self.repo.index.add([filename])
        self.repo.index.commit(filename)

    def _initialize_repo(
        self, repo_dir: pathlib.Path, initial_branch: str
    ) -> None:
        repo_dir.parent.mkdir(exist_ok=True, parents=True)
        self.repo = git.Repo.init(initial_branch=initial_branch, path=repo_dir)

    @abc.abstractmethod
    def __call__(self, repo_dir: pathlib.Path, initial_branch: str) -> None:
        return


class FreshClone(RepoSupport):
    def __init__(self, this_mocker, repo_instance) -> None:
        super().__init__()

        self.mocker = this_mocker
        self.repo_instance = repo_instance

        self._to_path = None
        self._git_clone = None

    def _file_clone(self, *args, **kwargs) -> git.Repo:
        return self._git_clone(
            f"file://{self.repo_instance.repo_dir.resolve()}", self._to_path
        )

    def __call__(self, repo_dir: pathlib.Path, initial_branch: str) -> None:
        # DO NOT initialise a git repo
        assert not self.repo
        assert not repo_dir.exists()

        self._to_path = repo_dir
        # preserve the original clone function so it can be used later
        self._git_clone = git.Repo.clone_from
        self.mocker.patch.object(
            pygitsync._repo.git.Repo, "clone_from", self._file_clone
        )


class DefaultBranchHead(RepoSupport):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, repo_dir: pathlib.Path, initial_branch: str) -> None:
        super()._initialize_repo(repo_dir, initial_branch)

        wtd = pathlib.Path(self.repo.working_tree_dir)
        self._commit_file(wtd / "c1")
        self._commit_file(wtd / "c2")


class TaggedRepo(RepoSupport):
    def __init__(self, tag_sequence: typing.List[str]) -> None:
        super().__init__()

        self.tag_sequence = tag_sequence

    def __call__(self, repo_dir: pathlib.Path, initial_branch: str) -> None:
        super()._initialize_repo(repo_dir, initial_branch)

        for x in range(len(self.tag_sequence)):
            self._commit_file(
                pathlib.Path(self.repo.working_tree_dir) / f"c{x}"
            )
            self.repo.create_tag(
                self.tag_sequence[x], message="tag at tag sequence {x}"
            )


class NonDefaultBranchHead(RepoSupport):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, repo_dir: pathlib.Path, initial_branch: str) -> None:
        super()._initialize_repo(repo_dir, initial_branch)

        wtd = pathlib.Path(self.repo.working_tree_dir)
        self._commit_file(wtd / "dc1")
        self._commit_file(wtd / "dc2")

        nb = self.repo.create_head("feature")
        self.repo.head.reference = nb
        self.repo.head.reset(index=True, working_tree=True)

        assert self.repo.active_branch.name == "feature"

        self._commit_file(wtd / "fc1")


T = typing.TypeVar("T", bound="MockRepo")


class MockRepo:
    """Manage test context of mock repo."""

    def __init__(
        self, repo_url: str, repo_support: RepoSupport, initial_branch: str
    ) -> None:
        self.initial_branch = initial_branch
        self.repo_url = repo_url
        self.repo_support = repo_support

        self.repo_dir = None
        self.working_dir = None

    @property
    def repo(self) -> typing.Optional[git.Repo]:
        return self.repo_support.repo

    @contextlib.contextmanager
    def instance(self) -> typing.Generator[T, None, None]:
        with tempfile.TemporaryDirectory() as d:
            self.working_dir = pathlib.Path(d)
            self.repo_dir = self.working_dir / extract_git_project(
                self.repo_url
            )

            # call the support for repo initialization
            self.repo_support(self.repo_dir, self.initial_branch)

            yield self



import pytest

from pygitsync._url import extract_git_project


class TestExtractGitProject:
    def test_http_clean(self):
        result = extract_git_project(
            "http://some.where/this/is/the/http_project"
        )

        assert result == "http_project"

    def test_http_git_suffix_stripped(self):
        result = extract_git_project(
            "http://some.where/this/is/the/http_project.git"
        )

        assert result == "http_project"

    def test_ssh_clean(self):
        result = extract_git_project("git@some.where:this/is/the/ssh_project")

        assert result == "ssh_project"

    def test_ssh_git_suffix_stripped(self):
        result = extract_git_project(
            "git@some.where:this/is/the/ssh_project.git"
        )

        assert result == "ssh_project"

    def test_raw_path_clean(self):
        result = extract_git_project("this/is/the/local_project")

        assert result == "local_project"

    def test_raw_path_suffix_stripped(self):
        result = extract_git_project("this/is/the/local_project.git")

        assert result == "local_project"



"""Entry to the main execution path."""

import asyncio
import logging
import pathlib
import tempfile
import typing

import click
import daemon  # type: ignore

from ._configuration import DEFAULT_CONFIGURATION_FILE, _load_configuration
from ._repo import _fetch_repo

log = logging.getLogger(__name__)


T = typing.TypeVar("T", bound="WhileContext")


class WhileContext:
    """Define exit status of the consuming while loop.

    If ``is_daemon`` is true, then the ``keep_running`` method always returns
    true. If not, then ``keep_running`` return true the first call, and then
    false after that.
    """

    def __init__(self: T, is_daemon: bool) -> None:
        """Construct ``WhileContext`` object.

        Args:
            is_daemon: Flag if the system is supposed to operate as a daemon.
        """
        self._continue = True
        self.is_daemon = is_daemon

    def keep_running(self: T) -> bool:
        """Define the run state of the caller.

        Returns:
            True for the consumer to continue running, False for the consumer
            to stop running.
        """
        if self.is_daemon:
            # always keep running
            return True
        if (not self.is_daemon) and self._continue:
            # let the loop run this time, but subsequent calls indicate exit
            self._continue = False
            return True
        else:
            # stop running
            return False


async def main(
    configuration_file: pathlib.Path,
    is_daemon: bool,
    sleep_interval_seconds: typing.Optional[float],
) -> None:
    """
    Primary execution path of the utility.

    ``is_daemon`` true means the loop will never exit, otherwise the loop only
    runs once and then exits.

    Args:
        configuration_file: Path to ``pygitsync`` configuration file.
        is_daemon: Flag to prevent execution from exiting.
        sleep_interval_seconds: Duration of period between git fetches.
    """
    with tempfile.TemporaryDirectory() as d:
        working_directory = pathlib.Path(d)

        this_configuration = await _load_configuration(
            configuration_file,
            is_daemon,
            sleep_interval_seconds,
        )

        loop = WhileContext(this_configuration.application.is_daemon)
        while loop.keep_running():
            try:
                log.debug("git sync execution placeholder")
                await _fetch_repo(this_configuration.repo, working_directory)

                log.debug(
                    "sleeping for, "
                    f"{this_configuration.application.sleep_interval_seconds} "
                    "seconds"
                )
                await asyncio.sleep(
                    this_configuration.application.sleep_interval_seconds
                )
            except Exception as e:
                log.critical(f"unhandled exception detected, {str(e)}")
                # sleep after logging this exception to avoid consuming CPU
                # if exception occurs very early in the loop (before the loop
                # hits its own sleep delay).
                await asyncio.sleep(
                    this_configuration.application.exception_sleep_seconds
                )


@click.command()
@click.option(
    "--configuration",
    "-c",
    "configuration_file",
    default=DEFAULT_CONFIGURATION_FILE,
    help="Path to pygitsync configuration file.",
    show_default=True,
    type=click.Path(path_type=pathlib.Path),
)
@click.option(
    "--daemon",
    "-d",
    "is_daemon",
    default=False,
    help="Run utility indefinitely in the background. "
    "Default: run utility indefinitely in foreground.",
    is_flag=True,
)
@click.option(
    "--interval",
    "-i",
    "sleep_interval_seconds",
    default=None,
    help="Duration in seconds between git fetches.",
    show_default=True,
    type=float,
)
def process_cli_arguments(
    configuration_file: pathlib.Path,
    is_daemon: bool,
    sleep_interval_seconds: typing.Optional[float],
) -> None:
    """Process CLI options."""
    if is_daemon:
        with daemon.DaemonContext():
            asyncio.run(
                main(configuration_file, is_daemon, sleep_interval_seconds)
            )
    else:
        # run with the daemon context.
        asyncio.run(main(configuration_file, is_daemon, sleep_interval_seconds))


def flit_entry() -> None:
    """Flit script entry point."""
    process_cli_arguments()



"""Manage git repository fetching."""

import logging
import pathlib
import re
import typing

import git

# NOTE: mypy complaining about missing type hints
import parver  # type: ignore

from ._configuration import GitRefType, RepoConfiguration
from ._url import extract_git_project

log = logging.getLogger(__name__)


class InvalidBranchError(Exception):
    """Problem occurred processing branch."""


def _get_remote(repo: git.Repo) -> git.Remote:
    remotes = repo.remotes
    if len(remotes) != 1:
        raise RuntimeError(
            "Something is wrong as there is more than one remote"
        )

    origin = remotes[0]

    return origin


def _evaluate_branch(name_pattern: str, repo: git.Repo) -> None:
    try:
        # name "pattern" must be an exact match to a single branch.
        if name_pattern != repo.active_branch.name:
            repo.heads[name_pattern].checkout()

        origin = _get_remote(repo)
        origin.pull()
    except Exception as e:
        raise InvalidBranchError(f"Invalid branch name, {name_pattern}") from e


def _evaluate_tag(name_pattern: str, repo: git.Repo) -> None:
    origin = _get_remote(repo)
    origin.fetch()

    pattern_matched_tags = [
        x for x in repo.tags if re.match(name_pattern, x.name)
    ]
    selected_tag: git.Reference
    if len(pattern_matched_tags) == 1:
        selected_tag = pattern_matched_tags[0]
        log.debug(
            f"single tag found matching pattern, {selected_tag.name} (tag), "
            f"{name_pattern} (pattern)"
        )
    elif len(pattern_matched_tags) != 0:
        log.debug(
            f"multiple tags found matching pattern, {name_pattern} (pattern), "
            f"{str([x.name for x in pattern_matched_tags])}"
        )
        sorted_tags = sorted(
            pattern_matched_tags,
            key=lambda x: parver.Version.parse(x.name),
            reverse=True,
        )
        selected_tag = sorted_tags[0]
        log.debug(f"selected tag from multiple, {selected_tag.name}")
    else:
        log.warning(f"No tags found matching pattern, {name_pattern}")
        return

    repo.head.reference = selected_tag  # type: ignore
    repo.head.reset(index=True, working_tree=True)


GIT_REF_METHOD: typing.Dict[
    GitRefType, typing.Callable[[str, git.Repo], None]
] = {
    GitRefType.branch: _evaluate_branch,
    GitRefType.tag: _evaluate_tag,
}


async def _fetch_repo(
    configuration: RepoConfiguration, working_dir: pathlib.Path
) -> None:
    project_name = extract_git_project(configuration.url)
    project_path = working_dir / project_name
    if not project_path.is_dir():
        log.debug(
            f"project directory does not exist. cloning a the repo from "
            f"configuration, {configuration.url}"
        )
        this_repo = git.Repo.clone_from(configuration.url, working_dir)
        log.info(
            f"Default active branch from fresh clone is, "
            f"{this_repo.active_branch.name}"
        )
    else:
        log.debug(f"existing project directory found, {project_path}")
        this_repo = git.Repo.init(project_path)

    GIT_REF_METHOD[configuration.pattern_type](configuration.pattern, this_repo)
